import time
from threading import Thread, Lock
from queue import Queue
from .utils import prepare_rich_logger, TaskStatus
from .environments import QueueBlockTime
import time

logger = prepare_rich_logger("Executor")


RunQueue = Queue()
CheckQueue = Queue()

# Ajust task counts if needed
lock = Lock()
ToAjustTaskCount = False  #
AjustTaskCount = 0


class RunTaskExecutor(Thread):
    def __init__(self, job):
        Thread.__init__(self)
        self._job = job

    def run(self) -> None:
        global ToAjustTaskCount, AjustTaskCount
        while True:
            logger.info("New run round begains")
            try:
                task = RunQueue.get(
                    block=True, timeout=QueueBlockTime
                )  # block untile timeout
            except:
                logger.warn("Empty RunQueue ")
                continue

            if task is None:
                break

            # logger.debug(f"Task name: {task.name}, task_id: {task.task_id}")
            predecessors = self._job.get_predecessors(task.id)
            task_predecessors_success_num = 0

            for predecessor in predecessors:
                # log here, and base the job
                status = predecessor.check_run_status(self._job._task_manager)
                if status == TaskStatus.SUCCESS:
                    task_predecessors_success_num += 1
            
            # print(50, task.id, len(predecessors), task_predecessors_success_num)
            if len(predecessors) == task_predecessors_success_num:
                try:
                    task.run(self._job._task_manager)  # mark failed
                except Exception as e:
                    logger.error(
                        f"Task `{task.name}` not been triggered successfully, details: {e.args}"
                    )
                    with lock:
                        ToAjustTaskCount = True
                        AjustTaskCount += self._job.get_descendant_counts(task.id)
                    continue
            else:
                RunQueue.put(task)
                time.sleep(QueueBlockTime)


class CheckTaskExecutor(Thread):
    def __init__(self, job, root_task_ids):
        Thread.__init__(self)
        self._job = job
        self._total_task_counts = len(job._tasks)  # UPDATE
        self._seen_tasks = set()
        for task_id in root_task_ids:
            self._seen_tasks.add(task_id)
        self._job_start_at = int(time.time())

    def run(self) -> None:
        global ToAjustTaskCount, AjustTaskCount
        success_checked = 0
        non_success_checked = 0

        while True:
            logger.info("New check round begains")
            # print(82, success_checked + non_success_checked,  self._total_task_counts)

            if (success_checked + non_success_checked) == self._total_task_counts:
                logger.info(f"All tasks of {self._job.name} finished")
                RunQueue.put(None)  # also notify RunQueue to terminate
                break

            if ToAjustTaskCount:
                with lock:
                    self._total_task_counts -= AjustTaskCount
                    AjustTaskCount = 0
                    ToAjustTaskCount = False

            try:
                task = CheckQueue.get(block=True, timeout=QueueBlockTime)
            except:
                logger.warn("Empty CheckQueue")
                continue

            status = task.check_run_status(self._job._task_manager)

            try:
                task.record(
                    self._job.name, self._job_start_at, status, int(time.time())
                )
            except AttributeError:
                pass

            if status in [TaskStatus.RUNNING, TaskStatus.PENDING]:
                CheckQueue.put(task)
                time.sleep(QueueBlockTime)

            elif status == TaskStatus.FAILED:
                non_success_checked += 1
                logger.error(f"Task `{task.name}` failed")
                with lock:
                    ToAjustTaskCount = True
                    AjustTaskCount += self._job.get_descendant_counts(task.id)
                continue

            elif status == TaskStatus.SUCCESS:
                success_checked += 1  

                successors = self._job.get_successors(task.id)

                for successor in successors:
                    if (
                        successor.id in self._seen_tasks
                    ):  # prevent running twice for same task
                        continue
                    self.put_task_and_record(successor)
                    CheckQueue.put(successor)

                # logger.info(f"Task `{task.id}` finished successfully")

    def put_task_and_record(self, task):
        RunQueue.put(task)
        self._seen_tasks.add(task.id)



import os
import time
import logging
import networkx as nx
import matplotlib.pyplot as plt
import platform
import subprocess
import redis 
import pickle
from enum import Enum
from rich import print as rprint
from rich.logging import RichHandler
from .exceptions import PyDagException
from .environments import ADD_SUDO_BOOL, HOST, PASSWORD, PREFIX

class TaskStatus(Enum):
    RUNNING = "running"
    FAILED = "failed"
    SUCCESS = "success"
    PENDING = "pending"


def draw_graph(g, name):
    options = {
        "font_size": 10,
        "node_size": 2000,
        "node_color": "#ff8c00",
        "edgecolors": "black",
        "linewidths": 1,
        "width": 2,
        "with_labels": True,
        "alpha": 0.8,
    }
    plt.figure(1, figsize=(10, 10))
    pos = nx.nx_pydot.pydot_layout(g, prog="dot")
    nx.draw(g, pos, **options)
    plt.savefig(f"{name}.png")


def prepare_rich_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    # create console handler and set level to debug
    rh = RichHandler()
    rh.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter("%(message)s")

    # add formatter to rh
    rh.setFormatter(formatter)

    # add rh to logger
    logger.addHandler(rh)

    return logger


def compose_task_name(job_name: str, task_name: str) -> str:
    name = f"{job_name}:{task_name}"
    if len(name) > 32:
        raise ValueError(
            "Task name too long! The length of `job_name` plus the length of `task_name` must within `31`"
        )
    return name


def compose_command(file: str, default_command_map={}, add_sudo=True) -> str:
    file_abpath = os.path.join(os.path.abspath(os.path.curdir), file)

    if not os.path.exists(file_abpath):
        raise PyDagException(f"file `{file}` not exists in curent directory")

    try:
        suffix = file.split(".")[-1]
    except:
        raise PyDagException(
            f"Wrong file name `{file}`, proper file name should have suffix, e.g, `.py`, `.ipynb`, etc"
        )
    
    if default_command_map != {}:
        try:
            command_for_file = default_command_map[suffix]
        except KeyError:
            raise PyDagException(f"No command for file type:`{suffix}` if found")
    else:
        command_for_file = get_command_by_suffix(suffix, add_sudo)

    if callable(command_for_file):
        return command_for_file(file_abpath)
    else:
        raise PyDagException(f"Command for file type:`{suffix}` is not callable, command must be callable with one argument `file`")

def compose_command_for_job(file, to_run_new: bool, add_sudo:bool) -> str:
    if to_run_new:
        return (
            get_enviroment_set_command_by_platform("PYDAG_RUN_NEW", "yes")
            + " && "
            + compose_command(file, add_sudo=add_sudo)
        )
    else:
        return (
            get_enviroment_set_command_by_platform("PYDAG_RUN_NEW", "no")
            + " && "
            + compose_command(file, add_sudo=add_sudo)
        )


def get_command_by_suffix(suffix: str, add_sudo):  # return a function, take file as its parameter
    if suffix == "py":
        command = os.environ.get("PYDAG_PYTHON_COMMAND", "")
        if command == "":
            command =  get_default_executable("python", add_sudo=add_sudo)  
        return lambda file: command + " " + file

    # Jupyter notebook is little bit different
    # command neet to be a function, take file as its parameter
    elif suffix == "ipynb":
        command = os.environ.get("PYDAG_JUPYTERNB_COMMAND", "")
        if command == "":
            command =  get_default_jupyternb_executable("jupyter", add_sudo=ADD_SUDO_BOOL)
        return lambda file: command + " " + file + " --stdout"

    else:
        raise PyDagException(f"File with subfix of `{suffix}` is not supported yet")


def get_default_executable(command: str, add_sudo=False):
    platform_name = platform.system()
    if platform_name == "Windows":
        executable = subprocess.check_output(["where", command]).decode("utf-8").strip()
        print(executable)
    elif platform_name == "Linux":
        executable = subprocess.check_output(["which", command]).decode("utf-8").strip()
    else:
        raise PyDagException(f"Platform `{platform_name}` is not supported yet")

    if add_sudo:
        return "sudo" + " " + executable
    else:
        return executable


def get_default_jupyternb_executable(command: str, add_sudo=False):
    platform_name = platform.system()
    if platform_name == "Windows":
        executable = subprocess.check_output(["where", command]).decode("utf-8").strip()
    elif platform_name == "Linux":
        executable = subprocess.check_output(["which", command]).decode("utf-8").strip()
    else:
        raise PyDagException(f"Platform `{platform_name}` is not supported yet")

    executable_jn = executable + " nbconvert --execute --to notebook"
    if add_sudo:
        return "sudo" + " " + executable_jn
    else:
        return executable_jn


def get_enviroment_set_command_by_platform(varible: str, value):
    platform_name = platform.system()
    if platform_name == "Windows":
        return f"set {varible}={value}"
    elif platform_name == "Linux":
        return f"export {varible}={value}"
    else:
        raise PyDagException(f"Platform `{platform_name}` is not supported yet")


def timestamp_to_datetime(job_run_at: str) -> str:
    return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(int(job_run_at)))


def list_pipelines(job_name=None):
    _store = redis.Redis(host=HOST, password=PASSWORD, db=3)
    if job_name == None:
        pattern = PREFIX + "*META"  # skip meta
        job_keys = _store.keys(pattern=pattern)

        jobs = []

        for job_key in job_keys:
            name = job_key.decode().split(":")[1]
            jobs.append(name)
        print("\n".join(jobs))
        return jobs 
    else:
        job_meta = _store.get(":".join([PREFIX, job_name, "META"]))

        if not job_meta:
            rprint(f"[red b]\[Pydag]There are no job named `{job_name}`")
            return

        job_graph = pickle.loads(job_meta)
        tasks = job_graph.nodes
        print("\n".join(tasks))
        return tasks


import redis
import pickle
from rich.tree import Tree
from rich import print as rprint
import pandas as pd
from dataclasses import dataclass
from networkx import DiGraph
from pygocron.pygocron import PyGoCron
from .environments import HOST, PASSWORD, PREFIX
from .utils import timestamp_to_datetime
from .exceptions import PyDagException


@dataclass
class TaskRecord:
    job_name: str
    job_run_at: int
    task_name: str
    run_id: int
    task_id: int
    status: str
    record_at: int

    def compose_key(self):
        return ":".join([PREFIX, self.job_name, str(self.job_run_at), self.task_name])

    def compose_value(self):
        return ",".join(
            [str(self.run_id), str(self.task_id), self.status, str(self.record_at)]
        )


class BasicJobLogger:
    def __init__(self):
        if HOST == "" or PASSWORD == "":
            raise PyDagException(
                f"""Please set `PYDAG_LOG_STORE_HOST` and `PYDAG_LOG_STORE_PASSWORD` envrionment varible,
                                `PYDAG_LOG_STORE_HOST` is a redis connection host and `PYDAG_LOG_STORE_PASSWORD`is a redis connection password"""
            )

        self._store = redis.Redis(host=HOST, password=PASSWORD, db=3)

    def record_job_info(self, job_name, dg: DiGraph):
        _key = ":".join([PREFIX, job_name, "META"])
        dg_picked = pickle.dumps(dg)
        if self._store.get(_key) != None:
            return

        self._store.set(_key, dg_picked)

    def record_task_info(
        self,
        job_name: str,
        job_run_at: int,
        task_name: str,
        run_id: int,
        task_id: int,
        status: str,
        task_record_at: int,
    ):
        record = TaskRecord(
            job_name, job_run_at, task_name, run_id, task_id, status, task_record_at
        )
        self._store.set(record.compose_key(), record.compose_value())

    def log_job(self, job_name, latest_n=3, style="tree"):
        pattern = ":".join([PREFIX, job_name]) + ":[1-9]*"  # skip meta
        task_keys = self._store.keys(pattern=pattern)
        job_meta = self._store.get(":".join([PREFIX, job_name, "META"]))
        if not job_meta:
            rprint(f"[red b]\[Pydag]There are no job named `{job_name}`")
            return

        job_graph = pickle.loads(job_meta)

        records = []

        for task in task_keys:
            keys = task.decode().split(":")
            values = self._store.get(task).decode().split(",")
            records.append(keys + values)

        record_df = pd.DataFrame(records)
        if record_df.empty:
            rprint(f"[red b]\[Pydag]No logs for job `{job_name}`")
            return
            
        record_df = pd.DataFrame(records).loc[:, 1:]
        record_df.columns = [
            "job_name",
            "job_run_at",
            "task_name",
            "run_id",
            "task_id",
            "status",
            "task_record_at",
        ]
        record_df = record_df.sort_values(
            ["job_run_at", "task_record_at"], ascending=[False, True]
        )
        job_run_ats = record_df["job_run_at"].unique()[:latest_n]
        record_df = record_df[record_df["job_run_at"].isin(job_run_ats)]
        log_groups = record_df.groupby("job_run_at")
        log_groups = sorted(log_groups, reverse=True)

        print_log_title(f"Log For Job `{job_name}`")
        for n, (job_run_at, df) in enumerate(log_groups):
            task_records = df.to_dict(orient="records")
            display_job_run(job_name, job_run_at, task_records, job_graph, method=style)
            if n != len(log_groups) - 1:
                print_log_sep()
        print_log_end()

    def log_task(self, job_name, task_name, latest_n=1):
        pyg = PyGoCron()
        try:
            task_id = pyg.get_task_id_by_name(":".join([job_name, task_name]))
        except:
            rprint(f"[red b]Error: Can not find the task `{task_name}`")
            return
        task_log = pyg.get_task_logs(task_id=task_id, page_size=latest_n)
        total = task_log["total"]
        print_log_title(f"Log For Task `{task_name}`")
        if total >= 1:
            logs = task_log["data"]
            for n, log in enumerate(logs):
                print(
                    f"`{log['name']}` at {log['end_time'].split('+')[0].replace('T', ' ')}"
                )
                print(log["result"].strip())
                if n != len(logs) - 1:
                    print_log_sep()
        print_log_end()

    def clear_log(self, job_name):
        """
        Clear all log and job meta info related to the job
        """

        def delete_keys(items):
            pipeline = self._store.pipeline()
            for item in items:
                pipeline.delete(item)
            pipeline.execute()

        pattern = ":".join([PREFIX, job_name]) + "*"
        task_keys = self._store.keys(pattern=pattern)
        delete_keys(task_keys)


def display_job_run(job_name, job_run_at, tasks, job_graph, method="flat"):
    job_run_at = timestamp_to_datetime(job_run_at)
    tree = Tree(f"`{job_name}` at {job_run_at}")

    root_task_names = [n for n, d in job_graph.in_degree() if d == 0]

    build_log_tree(tree, job_graph, root_task_names, tasks, method)

    rprint(tree)


def build_log_tree(tree, job_graph, root_task_names, tasks, method):
    if method == "tree":
        for root_task_name in root_task_names:
            build_dependant_log_tree(tree, job_graph, root_task_name, tasks)
    else:
        build_flatten_log_tree(tree, tasks)


def build_dependant_log_tree(tree, graph, root, tasks):
    try:
        task_info = [task for task in tasks if task["task_name"] == root][
            0
        ]
    except IndexError:
        return
    status = task_info["status"]
    task_name = task_info["task_name"]
    task_reocrd_at = timestamp_to_datetime(task_info["task_record_at"])

    _info = f"[{pick_color_for_status(status)} b] `{task_name}` at {task_reocrd_at}"
    tree = tree.add(_info)  # root

    for successor in graph.successors(root):
        build_dependant_log_tree(tree, graph, successor, tasks)


def build_flatten_log_tree(tree, tasks):
    for task in tasks:
        status = task["status"]
        task_name = task["task_name"]
        task_reocrd_at = timestamp_to_datetime(task["task_record_at"])
        _task = f"[{pick_color_for_status(status)} b] `{task_name}` at {task_reocrd_at}"
        tree.add(_task)


def pick_color_for_status(status: str):
    if status == "success":
        return "green"
    elif status == "failed":
        return "red"
    elif status == "running":
        return "blue"
    elif status == "pending":
        return "yellow"
    return "black"


def print_log_title(title):
    rprint("[blue b]" + ">> " * 5 + title + "<< " * 5)


def print_log_sep():
    rprint("[blue b]" + "== " * 20)


def print_log_end():
    rprint("[blue b]" + ">> " * 20)



from .utils import compose_command as _compose_command

__all__ = ["compose_command_for_file"]

def compose_command_for_file(file:str, default_command_map={}) -> str:
    return _compose_command(file, default_command_map)


from typing import List
from abc import ABC, abstractclassmethod
import networkx as nx
from pygocron.pygocron import PyGoCron
from .task import Task
from .utils import draw_graph, prepare_rich_logger
from .executor import RunQueue, CheckQueue, RunTaskExecutor, CheckTaskExecutor
from .environments import TO_RUN_NEW
from .exceptions import PyDagException
from .log import BasicJobLogger

logger = prepare_rich_logger("Job")


class TaskManger(ABC):
    @abstractclassmethod
    def create_task():
        ...

    @abstractclassmethod
    def run_task():
        ...

    @abstractclassmethod
    def check_run_status():
        ...


class Job:
    def __init__(self, name: str, task_manager: TaskManger) -> None:
        self.id = self.name = name
        self.root_tasks = None
        self._tasks: List[Task] = []
        self._graph = nx.DiGraph()
        self._task_manager = task_manager
        self._job_logger = BasicJobLogger()

    def add_task(self, *tasks):
        self._tasks.extend(tasks)

        self._graph.add_nodes_from([task._get_node_presentation() for task in tasks])

        edges = []
        existing_task_names = [task.id for task in self._tasks]

        for task in tasks:
            task_edges = task._get_edges()
            for pair in task_edges:
                for node in pair:
                    if node not in existing_task_names:
                        raise ValueError(
                            f"`Task<{node}>` seems forgoted to be added to the job"
                        )
            edges.extend(task_edges)

        self._graph.add_edges_from(edges)

        if not nx.is_directed_acyclic_graph(self._graph):
            raise ValueError(
                f"""Can not build a directed acyclic graph for job:`{self.name}`, 
                                   please ensure there's no loop among your tasks"""
            )

        self.root_tasks = self.get_root_tasks()

    def get_task_by_id(self, id):
        for task in self._tasks:
            if task.id == id:
                return task
        raise ValueError(f"Cannnot find a task given the id: {id}")

    def get_root_tasks(self,):
        root_ids = [n for n, d in self._graph.in_degree() if d == 0]    # [0]
        
        task = []
        for root_id in root_ids:
            task.append(self.get_task_by_id(root_id))
        return task

    def get_successors(self, id):
        successors = list(self._graph.successors(id))
        tasks = []

        for task in self._tasks:
            if task.id in successors:
                tasks.append(task)
        return tasks  #

    def get_descendant_counts(self, id):
        descendant = nx.descendants(self._graph, id)
        return len(descendant)

    def get_predecessors(self, id):
        predecessors = list(self._graph.predecessors(id))
        tasks = []

        for task in self._tasks:
            if task.id in predecessors:
                tasks.append(task)
        return tasks


class GoCronJob(Job):
    """
    Build a job based on cron
    """

    def __init__(self, name: str, task_manager: PyGoCron) -> None:
        super().__init__(name, task_manager)
        self._check_job_not_exists_before()

    def run(self, mode):
        try:
            assert mode in ["submit", "execute"]
        except AssertionError:
            raise PyDagException(f"Wrong mode: `{mode}`, should be one of [submit, execute]")
        logger.info(f"Start to run in `{mode}` mode")
        self._submit_tasks()
        logger.info("All tasks have been submitted")
        
        if mode == "execute":
            self._run()
    
    def _check_job_not_exists_before(self):
        if TO_RUN_NEW == "yes":
            logger.info("Prepare to run a brand new job")
            tasks = self._task_manager.get_tasks(
                tag=self.name
            )  # {'data': [], 'total': 0}
            if tasks["total"] > 0:
                raise PyDagException(
                    f"Job `{self.name}` already exists, please try another name, or you can delete the existing job and run again"
                )

        elif TO_RUN_NEW == "no":
            logger.info("Prepare to run a existed job")
            return
        else:
            raise PyDagException(
                f"Wrong `PYDAG_RUN_NEW` value: {TO_RUN_NEW}, must be one of [`yes`, `no`] (case insensitive)"
            )

    def _submit_tasks(self, log_job_meta=True):
        for task in self._tasks:
            task.submit(
                add_job_name=True,
                job_name=self.name,
                tag=self.name,
                task_manager=self._task_manager,
            )
        if log_job_meta:
            self._job_logger.record_job_info(self.name, self._graph)

    def _run(self,):
        logger.info(f"Job `{self.id}` trggered")
        root_tasks = self.get_root_tasks()
        
        for task in root_tasks:
            RunQueue.put(task)
            CheckQueue.put(task)

        run_executor = RunTaskExecutor(job=self)
        check_executor = CheckTaskExecutor(job=self, root_task_ids=[task.id for task in root_tasks])

        run_executor.start()
        check_executor.start()

        run_executor.join()
        check_executor.join()

    def __repr__(self):
        draw_graph(self._graph, self.name)
        return f"<{self.name} with {len(self._tasks)} tasks>"

    @staticmethod
    def delete_job_tasks(task_manager: PyGoCron, job_name: str):
        task_manager.delete_task_by_tag(job_name)
        logger.info(
            f"All tasks belongs to job `{job_name}` have been deleted successfully"
        )



from pygocron.pygocron import PyGoCron
from plan import Plan
from argparse import ArgumentParser
from .utils import compose_command_for_job, prepare_rich_logger, list_pipelines
from .job import GoCronJob
from .log import BasicJobLogger

logger = prepare_rich_logger("Pydag")


def submit(args):
    """
    plan job run by cron via `plan`
    """
    file = args.file
    job_name = args.name

    if not job_name:
        raise ValueError(
            "No `job` name provided, please define one via `--name=[job name]` flag"
        )

    command = compose_command_for_job(
        file, to_run_new=False, add_sudo=False
    )  # always run a existing job
    cron = Plan(name=job_name)

    cron_expression = args.cron
    if not cron_expression:
        raise ValueError(
            "No cron expresion provides, please define one via `--cron=[cron expression]` flag"
        )
    cron.command(command, every=cron_expression)
    try:
        cron.run("update")  # this problem is due to python plan module, for a none-root user, first time trigger a `update` will lead to a error
    except:
        cron.run("write")


def delete(args):
    pyg = PyGoCron()
    job_name = args.job_name

    # delete all task instance
    GoCronJob.delete_job_tasks(pyg, job_name)

    # delete associated cron-job
    cron = Plan(name=job_name)
    cron.run("clear")

    # delete associate logs
    bl = BasicJobLogger()
    bl.clear_log(job_name)


def log(args):
    bl = BasicJobLogger()
    

    style = args.style

    if args.task_name:
        bl.log_task(args.job_name, args.task_name, args.n)
    else:
        bl.log_job(args.job_name, args.n, style)

    
def list(args):
    job_name = args.job_name
    _ = list_pipelines(job_name)

class Submit:
    @staticmethod
    def register_subcommand(subparser):
        submit_parser = subparser.add_parser(
            "submit",
            help="CLI tool to submit a `pydag` job;Before running this command, you have to make sure you had submitted the job before",
        )
        submit_parser.add_argument(
            "file",
            type=str,
            help="Executable file to run, be aware that the `file` must exist in the current directory",
        )
        submit_parser.add_argument(
            "--cron",
            type=str,
            help="Cron expression for the job, then job will run according to that expression",
        )
        submit_parser.add_argument(
            "--name",
            type=str,
            help="Job name, this is optional, if not given, the `file` will treates as a job name; Not this job name can differ from the job name you defined in the file",
        )
        submit_parser.set_defaults(func=submit)


class Log:
    @staticmethod
    def register_subcommand(subparser):
        log_parser = subparser.add_parser(
            "log", help="CLI tool to get `pydag` job & task logs"
        )
        log_parser.add_argument("job_name", type=str, help="Job name")
        log_parser.add_argument(
            "--task_name",
            type=str,
            help="Task name, optional; If it's given, then the log for specific task will be shown",
        )
        log_parser.add_argument(
            "-n",
            type=int,
            default=3,
            help="The number of recent job logs, default is `3`",
        )
        log_parser.add_argument(
            "--style",
            type=str,
            default="flat",
            choices=["flat", "tree"],
            help="Print the log in `tree` format or `flat` format"
        )
        log_parser.set_defaults(func=log)


class Delete:
    @staticmethod
    def register_subcommand(subparser):
        delete_parser = subparser.add_parser(
            "delete", help="CLI tool to delete `pydag` job"
        )
        delete_parser.add_argument(
            "job_name",
            type=str,
            help="Job name, all tasks associate with the job name then will be deleted",
        )
        delete_parser.set_defaults(func=delete)

class List:
    @staticmethod
    def register_subcommand(subparser):
        list_parser = subparser.add_parser(
            "list", help="CLI tool to list all the `pydag` job and tasks"
        )
        list_parser.add_argument(
            "--job_name",
            type=str,
            help="Job name, all tasks associate with the job name then will be list",
        )
        list_parser.set_defaults(func=list)

def main():
    parser = ArgumentParser()
    subparser = parser.add_subparsers()
    Submit.register_subcommand(subparser)
    Log.register_subcommand(subparser)
    Delete.register_subcommand(subparser)
    List.register_subcommand(subparser)

    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.print_help()
        exit(1)

    args.func(args)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["pydot", "rich", "pygocron>=0.2.0", "pandas", "redis"]

test_requirements = []

setup(
    author="Jiang Chaodi",
    author_email="929760274@qq.com",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="lightweigt python workflow(pipeline) builder",
    entry_points={"console_scripts": ["pydag=pydag.cli:main",],},
    install_requires=requirements,
    license="MIT license",
    # long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="pydag2",
    name="pydag2",
    packages=find_packages(include=["pydag", "pydag.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/superjcd/pydag2",
    version="0.1.8",
    zip_safe=False,
)



import os
from .exceptions import PyDagException

# Meta 
PREFIX = os.environ.get("PYDAG_LOG_PREFIX", "Pydag")
QueueBlockTime = os.environ.get("PYDAG_ROUNDS_TIME", "10")
QueueBlockTime = int(QueueBlockTime)

# Log Store
HOST = os.environ.get("PYDAG_LOG_STORE_HOST", "")
PASSWORD = os.environ.get("PYDAG_LOG_STORE_PASSWORD", "")


# Other Settings
TO_RUN_NEW = os.environ.get("PYDAG_RUN_NEW", "yes").lower().strip()
TASK_TIMEOUT = int(os.environ.get("PYDAG_TASK_TIMEOUT", 3600))

ADD_SUDO = str(os.environ.get("PYDAG_ADD_SUDO", "no")) 
ADD_SUDO_BOOL = True

if ADD_SUDO.lower() != "yes":
    ADD_SUDO_BOOL = False


from pygocron.pygocron import RunStatus
from .utils import compose_task_name, TaskStatus
from .log import BasicJobLogger
from .environments import TO_RUN_NEW, TASK_TIMEOUT
from .exceptions import PyDagException


class Task:
    def __init__(self, name: str, command: str):
        self.id = self.name = name
        self.command = command
        self.upstreams = []
        self.downstreams = []

    def set_downstream(self, task: "Task"):
        if task not in self.downstreams:
            self.downstreams.append(task)

    def set_upstream(self, task: "Task"):
        if task not in self.upstreams:
            self.upstreams.append(task)

    def _get_node_presentation(self):
        return (self.id, self._get_infos())

    def _get_infos(self):
        return {"task_name": self.name, "command": self.command}

    def _get_edges(self):
        edges = []
        for successor in self.downstreams:
            edges.append((self.id, successor.id))
        for predecessor in self.upstreams:
            edges.append((predecessor.id, self.id))
        return edges

    def __repr__(self) -> str:
        return f"Task<{self.id}>"


class GoCronTask(Task):
    def __init__(self, name: str, command: str, node_id:int=None):
        super().__init__(name, command)
        self._task_logger = BasicJobLogger()
        self._node_id = node_id if node_id else 1

    def submit(self, add_job_name=False, job_name="", tag="", task_manager=None):
        """
        submit ad task
        """
        if task_manager == None:
            raise ValueError(
                "It seems you didn't provide a `task_manger` when definnig a `job`"
            )

        if add_job_name:
            if job_name:
                task_name = compose_task_name(job_name, self.name)
            else:
                raise ValueError(
                    "if `add_job_name` is `True`, then you must set a `job_name`"
                )
        else:
            task_name = self.name

        # check if is to run a brand new job or existing job
        if TO_RUN_NEW == "yes":
            task_id = task_manager.create_task(
                name=task_name,
                spec=None,
                tag=tag,
                command=self.command,
                level=2,
                timeout=TASK_TIMEOUT,
                host_id=self._node_id 

            )
        elif TO_RUN_NEW == "no":
            task_id = task_manager.get_task_id_by_name(task_name)
        else:
            raise PyDagException(
                f"Wrong `PYDAG_RUN_NEW` value, must be one of [`yes`, `no`] (case insensitive)"
            )

        if task_id:
            self.task_id = task_id
        else:
            raise PyDagException("Task submittion failed")

    def run(self, task_manager):
        """
        run a task, and a run id of the task will be set
        """

        run_id = task_manager.run_task(self.task_id)
        if run_id:
            self.run_id = run_id
        else:
            raise PyDagException("Task run failed")

    def check_run_status(self, task_manager) -> str:
        """
        get a runnning task status, either one of ["sucess", "failed", "pendding", "running"]
        """
        if not hasattr(self, "run_id"):
            # the run_id is set by run method, if task doesn't has a run id, meanning task didn't ran yet
            return TaskStatus.PENDING

        status = task_manager.check_run_status(self.task_id, self.run_id)

        if status == RunStatus.SUCCESS:
            return TaskStatus.SUCCESS
        elif status == RunStatus.FAILED:
            return TaskStatus.FAILED
        elif status == RunStatus.RUNNING:
            return TaskStatus.RUNNING
        elif status == RunStatus.PENDING:
            return TaskStatus.PENDING
        raise PyDagException(f"Wrong status: {status}")

    def record(self, job_name, job_run_at, status: TaskStatus, task_record_at):
        self._task_logger.record_task_info(
            job_name,
            job_run_at,
            self.name,
            self.run_id,
            self.task_id,
            status.value,
            task_record_at,
        )
